{
  This is a modification of the demo in
  TRichView\Demos\DelphiUnicode\Assorted\CustomDraw\Callouts\

  Changes:
  1) When adding a right-side message, moving all text in it to the right of
    the cell (using MoveContentRight)
  2) Doing the same for all existing right-side messages on resizing
  3) DrawCallout draws a speech bubble around the cell content

  ---

  This demo uses undocumented property of RVData: collection of DrawItems.
  Each item in this collection represents some visible object in the document,
  such as a text line. It has properties: Left, Top, Width, Height.
  This collection is created when TRichView is formatted, so all functions
  accessing this collection are called when RichView1 is formatted.
}


unit MainFrm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Jpeg,
  Controls, Forms, Dialogs, PngImage, Math,
  RVStyle, RVScroll, RichView, RVEdit, RVTable, RVItem, CRVFData, DLines;

type

  TCalloutSide = (clsTopLeft, clsTopRight, clsBottomLeft, clsBottomRight,
    clsTop, clsBottom, clsLeft, clsRight, clsCallLeft, clsCallRight);

  {
    This class is used to draw a speech balloon
  }

  TCallout = class
  private
    FImages: array[TCalloutSide] of TGraphic;
    FColor: TColor;
    function GetBorderWidth: Integer;
    function GetBorderHeight: Integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadImages;
    procedure Draw(Canvas: TCanvas; const R: TRect; TalkFromLeft: Boolean);
    procedure DrawAround(Canvas: TCanvas; R: TRect; TalkFromLeft: Boolean);
    property BorderWidth: Integer read GetBorderWidth;
    property BorderHeight: Integer read GetBorderHeight;
    function GetLeftBorderWidth(TalkFromLeft: Boolean): Integer;
    function GetRightBorderWidth(TalkFromLeft: Boolean): Integer;
  end;
  {
    This class stores information about a chat visitor
  }

  TPerson = class
  private
    FImage: TGraphic;
    FName: String;
    FTextStyleNo: Integer;
  public
    constructor Create(ImageFileName, Name: String; TextStyleNo: Integer);
    destructor Destroy; override;
  end;


  TfrmMain = class(TForm)
    RVStyle1: TRVStyle;
    RichView1: TRichView;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RichView1Resize(Sender: TObject);
  private
    { Private declarations }
    FCallout: TCallout;
    FPerson1, FPerson2: TPerson;
  public
    { Public declarations }
    procedure AddMessage(Person: TPerson; TalkFromLeft: Boolean; const Msg: String);
    procedure DrawCallout(Sender: TRVTableItemInfo; Canvas: TCanvas;
      Left, Top, Right, Bottom: Integer;
      TableSelected, CellSelected, Printing: Boolean; Row, Col: Integer;
      var DoDefault: Boolean);
    procedure AdjustTextPosition;
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

const
  // Names of files represending images for drawing speech balloons
  CalloutFileNames: array[TCalloutSide] of String =
    ('tl', 'tr', 'bl', 'br', 't', 'b', 'l', 'r', 'cl', 'cr');
  // Indexes in RVStyle1.ParaStyles
  PARA_NORMAL = 0;
  PARA_CENTERED = 1;
  PARA_MESSAGE = 2;

{================================ TCallout ====================================}
constructor TCallout.Create;
var
  i: TCalloutSide;
begin
  inherited Create;
  FColor := clWhite;
  for i := Low(TCalloutSide) to High(TCalloutSide) do
    FImages[i] := TPngImage.Create;
end;
{------------------------------------------------------------------------------}
destructor TCallout.Destroy;
var
  i: TCalloutSide;
begin
  for i := Low(TCalloutSide) to High(TCalloutSide) do
    FImages[i].Free;
  inherited Destroy;
end;
{------------------------------------------------------------------------------}
procedure TCallout.LoadImages;
var
  i: TCalloutSide;
  Path: String;
begin
  Path := ExtractFilePath(Application.ExeName) + 'Graphic\';
  for i := Low(TCalloutSide) to High(TCalloutSide) do
    FImages[i].LoadFromFile(Path+CalloutFileNames[i]+'.png');
end;
{------------------------------------------------------------------------------}
procedure TCallout.Draw(Canvas: TCanvas; const R: TRect; TalkFromLeft: Boolean);
var
  InnerR: TRect;
  TL, TR: TGraphic;
  DL, DR: Integer;
begin
  if TalkFromLeft then
  begin
    TL := FImages[clsCallLeft];
    TR := FImages[clsTopRight];
  end
  else
  begin
    TL := FImages[clsTopLeft];
    TR := FImages[clsCallRight];
  end;
  DL := TL.Width - FImages[clsTopLeft].Width;
  DR := TR.Width - FImages[clsTopRight].Width;

  // Drawing inner rectangle
  Canvas.Brush.Color := FColor;
  Canvas.Brush.Style := bsSolid;
  InnerR.Left := R.Left + FImages[clsLeft].Width + DL;
  InnerR.Right := R.Right - FImages[clsRight].Width - DR;
  InnerR.Top := R.Top + FImages[clsTop].Height;
  InnerR.Bottom := R.Bottom - FImages[clsBottom].Height;
  Canvas.FillRect(InnerR);
  // Drawing sides
  Canvas.StretchDraw(
    Rect(R.Left + FImages[clsTopLeft].Width + DL, R.Top,
      R.Right - FImages[clsTopRight].Width - DR, R.Top + FImages[clsTop].Height),
    FImages[clsTop]);
  Canvas.StretchDraw(
    Rect(R.Left + FImages[clsBottomLeft].Width + DL,
      R.Bottom - FImages[clsBottom].Height,
      R.Right - FImages[clsBottomRight].Width - DR, R.Bottom),
    FImages[clsBottom]);
  Canvas.StretchDraw(
    Rect(R.Left + DL, R.Top + FImages[clsTopLeft].Height,
      R.Left + FImages[clsLeft].Width + DL, R.Bottom - FImages[clsBottomLeft].Height),
    FImages[clsLeft]);
  Canvas.StretchDraw(
    Rect(R.Right - FImages[clsTopRight].Width - DR,
     R.Top + FImages[clsTopRight].Height,
      R.Right - DR, R.Bottom - FImages[clsBottomRight].Height),
    FImages[clsRight]);
  // Drawing corners
  Canvas.Draw(R.Left, R.Top, TL);
  Canvas.Draw(R.Left + DL, R.Bottom - FImages[clsBottomLeft].Height,
    FImages[clsBottomLeft]);
  Canvas.Draw(R.Right - TR.Width, R.Top,
    TR);
  Canvas.Draw(R.Right - FImages[clsBottomRight].Width - DR,
    R.Bottom - FImages[clsBottomRight].Height,
    FImages[clsBottomRight]);

end;
{------------------------------------------------------------------------------}
procedure TCallout.DrawAround(Canvas: TCanvas; R: TRect; TalkFromLeft: Boolean);
begin
  dec(R.Top, GetBorderHeight);
  inc(R.Bottom, GetBorderHeight);
  dec(R.Left, GetLeftBorderWidth(TalkFromLeft));
  inc(R.Right, GetRightBorderWidth(TalkFromLeft));
  Draw(Canvas, R, TalkFromLeft);
end;
{------------------------------------------------------------------------------}
function TCallout.GetBorderHeight: Integer;
begin
  Result := Max(FImages[clsTop].Height, FImages[clsBottom].Height);
end;
{------------------------------------------------------------------------------}
function TCallout.GetBorderWidth: Integer;
begin
  Result :=
    Max(
      Max(FImages[clsBottomLeft].Width, FImages[clsBottomRight].Width),
      Max(FImages[clsTopLeft].Width, FImages[clsTopRight].Width));
  Result :=
    Max(
      Result,
      Max(FImages[clsCallLeft].Width, FImages[clsCallRight].Width));
end;
{------------------------------------------------------------------------------}
function TCallout.GetLeftBorderWidth(TalkFromLeft: Boolean): Integer;
begin
  if TalkFromLeft then
    Result := Max(
      Max(FImages[clsCallLeft].Width, FImages[clsBottomLeft].Width),
      FImages[clsLeft].Width)
  else
    Result := Max(
      Max(FImages[clsTopLeft].Width, FImages[clsBottomLeft].Width),
      FImages[clsLeft].Width);
end;
{------------------------------------------------------------------------------}
function TCallout.GetRightBorderWidth(TalkFromLeft: Boolean): Integer;
begin
  if TalkFromLeft then
    Result := Max(
      Max(FImages[clsTopRight].Width, FImages[clsBottomRight].Width),
      FImages[clsRight].Width)
  else
    Result := Max(
      Max(FImages[clsCallRight].Width, FImages[clsBottomRight].Width),
      FImages[clsRight].Width);
end;
{===================================== TPerson ================================}
constructor TPerson.Create(ImageFileName, Name: String; TextStyleNo: Integer);
begin
  FImage := TPngImage.Create;
  FImage.LoadFromFile(ExtractFilePath(Application.ExeName) + 'Graphic\' + ImageFileName);
  FName := Name;
  FTextStyleNo := TextStyleNo;
end;
{------------------------------------------------------------------------------}
destructor TPerson.Destroy;
begin
  FImage.Free;
  inherited;
end;
{==============================================================================}
{
  Returning a text style (index in RVStyle.TextStyles) to draw a person name.
}
function GetColoredText(RVStyle: TRVStyle; Color: TColor): Integer;
var
  TextStyle: TFontInfo;
begin
  TextStyle := TFontInfo.Create(nil);
  TextStyle.Assign(RVStyle.TextStyles[0]);
  TextStyle.Size := TextStyle.Size + 2;
  TextStyle.Color := Color;
  TextStyle.Style := [fsBold];
  Result := RVStyle.FindTextStyle(TextStyle);
end;
{------------------------------------------------------------------------------}
{ Returns a rectangle around the content.
  This procedure assumes at least one item. }
procedure GetContentRect(RVData: TCustomRVFormattedData; var R: TRect);
var
  i: Integer;
  DItem: TRVDrawLineInfo;
  R2: TRect;
begin
  DItem := RVData.DrawItems[0];
  R := Bounds(DItem.Left, DItem.Top, DItem.Width, DItem.Height);

  for i := 1 to RVData.DrawItems.Count - 1 do
  begin
    DItem := RVData.DrawItems[i];
    R2 := Bounds(DItem.Left, DItem.Top, DItem.Width, DItem.Height);
    R := Rect(Min(R.Left, R2.Left), Min(R.Top, R2.Top),
      Max(R.Right, R2.Right), Max(R.Bottom, R2.Bottom));
  end;
end;
{------------------------------------------------------------------------------}
procedure MoveContentRight(RVData: TCustomRVFormattedData; MinDistance: Integer);
var
  R: TRect;
  i, DX: Integer;
begin
  GetContentRect(RVData, R);
  DX := RVData.GetWidth - R.Right - MinDistance;
  if DX <= 0 then
    exit;
  for i := 0 to RVData.DrawItems.Count - 1 do
    RVData.DrawItems[i].Left := RVData.DrawItems[i].Left + DX;

end;
{============================== TfrmMain ======================================}
procedure TfrmMain.FormCreate(Sender: TObject);
var
  jpg: TJPEGImage;
begin
  // creating a class for drawing speech balloon
  FCallout := TCallout.Create;
  FCallout.LoadImages;
  // creating classes for chat visitors
  FPerson1 := TPerson.Create('Avatar1.png', 'Bob', GetColoredText(RVStyle1, $00D3311F));
  FPerson2 := TPerson.Create('Avatar2.png', 'Alice', GetColoredText(RVStyle1, $00004284));
  // loading a chat background picture
  jpg := TJPEGImage.Create;
  jpg.LoadFromFile(ExtractFilePath(Application.ExeName) + 'Graphic\Back.jpg');
  RichView1.BackgroundBitmap.Assign(jpg);
  jpg.Free;
  RichView1.BackgroundStyle := bsStretched;
  // assigning indents around the paragraphs used for messages
  // (they must be large enough to allow drawing a balloon)
  with RVStyle1.ParaStyles[PARA_MESSAGE] do
  begin
    SpaceBefore := FCallout.BorderHeight;
    SpaceAfter := SpaceBefore;
    LeftIndent := FCallout.GetBorderWidth;
    RightIndent := LeftIndent;
  end;
  // adding sample chat conversation
  AddMessage(FPerson1, True, 'Hello!');
  AddMessage(FPerson2, False, 'Hello!'#13'I am online!');
  AddMessage(FPerson1, True, 'Please talk as much as you can.');
  AddMessage(FPerson2, False, 'Why?');
  AddMessage(FPerson1, True, 'We need to fill space in this window.');
  AddMessage(FPerson2, False, 'Ok.'#13'I understand.'#13+
    'I can demonstrate that this bubble can be long.'#13+
    'Because actually it is a table cell.'#13+
    'End of chat');
end;
{------------------------------------------------------------------------------}
procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  FCallout.Free;
  FPerson1.Free;
  FPerson2.Free;
end;
{------------------------------------------------------------------------------}
{ Adjusting position of message text in all existing messages. }
procedure TfrmMain.AdjustTextPosition;
var
  i, r, c: Integer;
  Table: TRVTableItemInfo;
  D: Integer;
begin
  if FCallout = nil then
    exit;
  D := FCallout.GetRightBorderWidth(False);
  for i := 0 to RichView1.ItemCount - 1 do
    if RichView1.GetItem(i) is TRVTableItemInfo then
    begin
      Table := TRVTableItemInfo(RichView1.GetItem(i));
      for r := 0 to Table.RowCount - 1 do
        for c := 0 to Table.ColCount - 1 do
          if (Table.Cells[r, c] <> nil) and (Table.Cells[r, c].Tag = '0') then
            MoveContentRight(Table.Cells[r, c], D);
    end;
end;
{------------------------------------------------------------------------------}
{ OnResize event }
procedure TfrmMain.RichView1Resize(Sender: TObject);
begin
  AdjustTextPosition;
end;
{------------------------------------------------------------------------------}
{ Custom drawing of cell background }
procedure TfrmMain.DrawCallout(Sender: TRVTableItemInfo; Canvas: TCanvas;
  Left, Top, Right, Bottom: Integer;
  TableSelected, CellSelected, Printing: Boolean; Row, Col: Integer;
  var DoDefault: Boolean);
var
  R: TRect;
  Cell: TRVTableCellData;
begin
  if (Row >= 0) and (Col >= 0) and (Sender.Cells[Row, Col].Tag <> '') then
  begin
    Cell := Sender.Cells[Row, Col];
    if TableSelected or CellSelected then
    begin
      Canvas.Brush.Color := clHighlight;
      Canvas.Brush.Style := bsSolid;
      Canvas.FillRect(Rect(Left, Top, Right, Bottom));
    end;
    GetContentRect(Cell, R);
    OffsetRect(R, Left + Sender.CellHPadding, Top + Sender.CellVPadding);
    FCallout.DrawAround(Canvas, R, Cell.Tag <> '0');
    DoDefault := False;
  end;
end;
{------------------------------------------------------------------------------}
{ Adding a new message to the chat }
procedure TfrmMain.AddMessage(Person: TPerson; TalkFromLeft: Boolean;
  const Msg: String);
var
  Table: TRVTableItemInfo;
  PersonCell, MsgCell: TRVTableCellData;
begin
  Table := TRVTableItemInfo.CreateEx(1, 2, RichView1.RVData);
  Table.ParaNo := PARA_CENTERED;
  Table.Color := clNone;
  Table.OnDrawBackground := DrawCallout;
  Table.CellHPadding := 2;
  Table.CellVPadding := 2;
  Table.Rows[0].VAlign := rvcTop;
  if TalkFromLeft then
  begin
    PersonCell := Table.Cells[0, 0];
    MsgCell := Table.Cells[0, 1];
  end
  else
  begin
    PersonCell := Table.Cells[0, 1];
    MsgCell := Table.Cells[0, 0];
  end;
  PersonCell.BestWidth := 70;
  PersonCell.Clear;
  PersonCell.AddPictureEx('', Person.FImage, PARA_CENTERED, rvvaBaseLine);
  PersonCell.SetItemExtraIntProperty(PersonCell.ItemCount - 1, rvepShared, 1);
  PersonCell.AddNL(Person.FName, Person.FTextStyleNo, PARA_CENTERED);
  MsgCell.Clear;
  MsgCell.AddTextNLW(Msg, 0, PARA_MESSAGE, PARA_MESSAGE, True);
  MsgCell.Tag := IntToStr(Ord(TalkFromLeft));

  RichView1.AddItem('', Table);
  RichView1.FormatTail;
  if not TalkFromLeft then
    MoveContentRight(MsgCell, FCallout.GetRightBorderWidth(False));

end;


end.
